from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..client import Client


class Resource:

    client: Client

    def __init__(
        self,
        client: Client,
    ):
        self.client = client
